EXPORTS
    MediaInfo_New                  @26
    MediaInfo_New_Quick            @30
    MediaInfo_Delete               @27
    MediaInfo_Open                 @14
    MediaInfo_Open_Buffer          @15
    MediaInfo_Open_Buffer_Init     @64
    MediaInfo_Open_Buffer_Continue @65
    MediaInfo_Open_Buffer_Continue_GoTo_Get @66
    MediaInfo_Open_Buffer_Finalize @67
    MediaInfo_Open_NextPacket      @76
    MediaInfo_Save                 @16
    MediaInfo_Close                @17
    MediaInfo_Inform               @18
    MediaInfo_GetI                 @19
    MediaInfo_Get                  @20
    MediaInfo_SetI                 @21
    MediaInfo_Set                  @22
    MediaInfo_Output_Buffer_Get    @72
    MediaInfo_Output_Buffer_GetI   @74
    MediaInfo_Option               @23
    MediaInfo_State_Get            @24
    MediaInfo_Count_Get            @25
    MediaInfoA_New                 @28
    MediaInfoA_New_Quick           @31
    MediaInfoA_Delete              @29
    MediaInfoA_Open                @1
    MediaInfoA_Open_Buffer         @6
    MediaInfoA_Open_Buffer_Init    @68
    MediaInfoA_Open_Buffer_Continue @69
    MediaInfoA_Open_Buffer_Continue_GoTo_Get @70
    MediaInfoA_Open_Buffer_Finalize @71
    MediaInfoA_Open_NextPacket     @77
    MediaInfoA_Save                @10
    MediaInfoA_Close               @2
    MediaInfoA_Inform              @3
    MediaInfoA_GetI                @4
    MediaInfoA_Get                 @5
    MediaInfoA_SetI                @8
    MediaInfoA_Set                 @9
    MediaInfoA_Output_Buffer_Get   @73
    MediaInfoA_Output_Buffer_GetI  @75
    MediaInfoA_Option              @11
    MediaInfoA_State_Get           @12
    MediaInfoA_Count_Get           @7
    MediaInfo_Info_Version         @13
    MediaInfoList_New              @32
    MediaInfoList_New_Quick        @33
    MediaInfoList_Delete           @34
    MediaInfoList_Open             @35
    MediaInfoList_Open_Buffer      @36
    MediaInfoList_Save             @37
    MediaInfoList_Close            @38
    MediaInfoList_Inform           @39
    MediaInfoList_GetI             @40
    MediaInfoList_Get              @41
    MediaInfoList_SetI             @42
    MediaInfoList_Set              @43
    MediaInfoList_Option           @44
    MediaInfoList_State_Get        @45
    MediaInfoList_Count_Get        @46
    MediaInfoList_Count_Get_Files  @62
    MediaInfoListA_New             @47
    MediaInfoListA_New_Quick       @48
    MediaInfoListA_Delete          @49
    MediaInfoListA_Open            @50
    MediaInfoListA_Open_Buffer     @51
    MediaInfoListA_Save            @52
    MediaInfoListA_Close           @53
    MediaInfoListA_Inform          @54
    MediaInfoListA_GetI            @55
    MediaInfoListA_Get             @56
    MediaInfoListA_SetI            @57
    MediaInfoListA_Set             @58
    MediaInfoListA_Option          @59
    MediaInfoListA_State_Get       @60
    MediaInfoListA_Count_Get       @61
    MediaInfoListA_Count_Get_Files @63
    JNI_OnLoad                     @78
;Last item = 78
