##  Copyright (c) MediaArea.net SARL. All Rights Reserved.
 #
 #  Use of this source code is governed by a BSD-style license that can
 #  be found in the License.html file in the root of the source tree.
 ##

#! /bin/sh

#-----------------------------------------------------------------------
# Default script
. ./Release_DLL_GNU.sub

#-----------------------------------------------------------------------
# Launch
Release_DLL Linux i386 so so.0 so.0.0.0
